/*--------------------------------------------------------------------------
core.c:
This file contains the main PIC16F84 emulation code.
Anything inside the chip is emulated here.
--------------------------------------------------------------------------*/

#include <stdlib.h>
#include <stdio.h>
#include "pic.h"

#ifdef SOUND
#include "sound.c"
#endif

// PC register can be incremented or write/exchanged but NOT both.
// This flag signals a PC write, and, if true, PC MUST NOT be incremented
// It is set by executing GOTO, CALL, RETLW, RETURN, RETFIE, any
// PCL modification or interrupt call

int pcwr,dis;

// Reset: init registers to its initial values

void f84_reset(core *pcore,int type)
{
        if(type==POR) pcore->Clocks=0;        //Profiling Time counter
        pcore->sp=0;            //Initial stack level
        pcore->pc[0]=0;         //Execution start address
        pcore->ir=0;            //Previous fetch = NOP
        if(type==POR) pcore->status=0x18;
        else pcore->status&=0x18;
        pcore->pclatch=0;
        pcore->intcon=0;
        pcore->option=0xff;
        pcore->trisa=0x1f;
        pcore->trisb=0xff;
        pcore->eecon1=0;
        pcore->porta=0x0;         // Arbitrary low output
        pcore->portb=0x0;         // Arbitrary low output
        pcore->tmr0=0;
        pcore->sync=0;
        pcore->prescaler=0;
        pcore->wdt=0;
        pcore->sleep=0;
        pcore->intcon=0;
}

/*---------------------------------------------------
Register-File Read.
This function must be inlined for faster code.
---------------------------------------------------*/

extern inline byte rd_f(core *pcore, byte ff)
{
        byte f=ff;
        if ((ff&0x7f)>=0x50) goto noreg;
        if ((ff&0x7f)>=0x0c) return pcore->f[ff-0xc];
        if (pcore->status&RP0F) f|=0x80;
        switch(f) {
        case 0x01: return pcore->tmr0;
        case 0x81: return pcore->option;
        case 0x02:
        case 0x82: return pcore->pc[pcore->sp]&0xff;
        case 0x03:
        case 0x83: return pcore->status;
        case 0x04:
        case 0x84: return pcore->fsr;
        case 0x05: return (porta_rd(pcore) & pcore->trisa) |
                          (pcore->porta & (~pcore->trisa));
        case 0x85: return pcore->trisa;
        case 0x06: return (portb_rd(pcore) & pcore->trisb) |
                          (pcore->portb & (~pcore->trisb));
        case 0x86: return pcore->trisb;
        case 0x08: return pcore->eeprom[pcore->eeadr&63];
        case 0x88: return pcore->eecon1;
        case 0x09: return pcore->eeadr;
        case 0x0a:
        case 0x8a: return pcore->pclatch;
        case 0x0b:
        case 0x8b: return pcore->intcon;
        }
noreg:
#ifdef DEBUG
        fprintf(stderr,"***Invalid read register:0x%X\n",f);
#endif
        return 0;
}

/*------------------------------------------------------
Register-File Write.
This function must be inlined for faster code.
A lot of peripheral emulation is done here.
------------------------------------------------------*/
// EEprom stuff:
unsigned long tstamp55=0,tstampAA=0;
word eetim=0;
byte eeadr,eedata;

extern inline void wr_f(core *pcore, byte ff, byte data)
{
        byte f=ff;
        word i;
        if ((ff&0x7f)>=0x50) goto noreg;
        if ((ff&0x7f)>=0x0c) { pcore->f[ff-0xc]=data; return; }
        if (pcore->status&RP0F) f|=0x80;
        switch(f) {
        // Writes to TMR0 stops the counter during 2 cycles.
        // And resets the prescaler.
        case 0x01: pcore->tmr0=data; pcore->prescaler=0; pcore->sync=3;
                   return;
        case 0x81: pcore->option=data; return;
        case 0x02:
        // Writes to PCL also transfers PCLATCH to PC as MSB, and discards
        // the currently prefetched instruction.
        case 0x82: i=pcore->pclatch; i<<=8; i|=data; i&=0x1fff;
                   pcore->pc[pcore->sp]=i;
                   pcwr=1;       // No PC increment this cycle.
                   dis=1;
                   pcore->ir=0;  // Discard prefetched
                   return;
        case 0x03:
        case 0x83: pcore->status=data; return;
        case 0x04:
        case 0x84: pcore->fsr=data; return;
        case 0x05: pcore->porta=data; porta_wr(pcore); return;
        case 0x85: pcore->trisa=data; porta_wr(pcore); return;
        case 0x06: pcore->portb=data; portb_wr(pcore); return;
        case 0x86: pcore->trisb=data; portb_wr(pcore); return;
        // Eeprom write emulation. To unlock eeprom writes the
        // sequence 0x55,0xAA must be write to EECON2 (0x89).
        // Then the write operation must be enabled in EECON1 (0x88).
        // The nominal time for Eeprom writes is 4 ms.
        // At the end of Eeprom write a flag is set and and interrupt
        // can be posted. The end of eeprom write is handeld in the
        // f84_run function.
        case 0x08: pcore->eedatawr=data; return; // Eeprom write
        case 0x88: pcore->eecon1=data&0xfe;      //Eeprom control 1
                   if ((data&6)==6) {
                       if((pcore->Clocks-tstampAA)==1) {
                          eeadr=pcore->eeadr;
                          eedata=pcore->eedatawr;
                          eetim=4*(pcore->freq/1000);
                       } else {
                           pcore->eecon1&=~2; pcore->eecon1|=8; //error
                       }
                   }
                   return;
        case 0x09: pcore->eeadr=data; return;
        case 0x89: // EECON2: EEprom write unlock
                   if(data==0x55) {tstamp55=pcore->Clocks; return;}
                   if(data==0xaa && (pcore->Clocks-tstamp55)<=2)
                        tstampAA=pcore->Clocks;
                   return;
        case 0x0a:
        case 0x8a: pcore->pclatch=data; return;
        case 0x0b:
        case 0x8b: pcore->intcon=data; return;
        }
noreg:
#ifdef DEBUG
        fprintf(stderr,"***Invalid write register:0x%X\n",f);
#endif
        return;
}

//*******************************************************************
extern int debug_status,brk;
extern unsigned long tbrk;

void f84_run(core *pcore, word *prg,int nc)
{
        word ireg,lk;
        byte f,d,b,k,tmp;
        static byte otmr0=0,opr=0,opa,opb;
        int i,j,n;

        for(;nc;nc--) {       // Execution loop: NC cycles
                // Trace: Check for breakpoints
                if((pcore->pc[pcore->sp]==brk)||
                   (tbrk && (tbrk==pcore->Clocks))) debug_status=STEP;
                // Call debugger if needed
                if(debug_status!=RUNNING) debug(pcore,prg);

                pcwr=0; // By default PC is incremented each cycle.
                dis=0; // If prefetch discard flag

                // In the SLEEP condition no clock is provided to the CPU
                // But Watchdog, External interrupts and EEprom writes must
                // be checked.
                if (pcore->sleep) goto chkhw;

                // Fetch a new instruction (it will be executed next clock
                // cycle if no jump takes place)
                ireg=pcore->ir;
                pcore->ir=prg[(pcore->pc[pcore->sp])&1023]&0x3fff;
                // Execute current instruction
                // Extract operands fields
                lk=ireg&0x7ff; k=ireg&0xff;     // Literal operands
                f=ireg&0x7f; d=ireg&0x80;       // File-Register index & dest.
                if (!f) f=pcore->fsr;   //indirect addressing: index=FSR
                b=(ireg>>7)&7;          // Bit number for Bxxx instructions
                // Execution branch:
                switch(ireg>>11) {      // First check 3-bit OPcodes
                case 4:                             // CALL
                    pcore->sp=(pcore->sp+1)&7;      // new stack level
                    pcore->pc[pcore->sp]=lk;        // Goto literal
                    pcwr=1;                         // No PC incr.
                    dis=1;
                    pcore->ir=0;                    // Discard prefetched
                    break;
                case 5:                             // GOTO
                    pcore->pc[pcore->sp]=lk;        // Goto literal
                    pcwr=1;                         // No PC incr.
                    dis=1;
                    pcore->ir=0;                    // Discard prefetched
                    break;
                default:
                    switch(ireg>>10) {      // Check 4 bit Opcodes
                    case 0xc:                       // MOVLW
                        pcore->w=k; break;
                    case 0xd:                       // RETLW
                        pcore->w=k;
                        dis=1;
                        pcore->ir=0;                // Discard prefetched
                        pcore->sp=(pcore->sp-1)&7;  // old stack level
                        pcwr=1;                     // No PC incr.
                        break;
                    case 4:                         // BCF
                        tmp=rd_f(pcore,f);
                        tmp&=~(1<<b);
                        wr_f(pcore,f,tmp);
                        break;
                    case 5:                         // BSF
                        tmp=rd_f(pcore,f);
                        tmp|=(1<<b);
                        wr_f(pcore,f,tmp);
                        break;
                    case 6:                         // BTFSC
                        if ((rd_f(pcore,f)&(1<<b))==0)
                            {pcore->ir=0; dis=1;}   // Discard prefetched
                        break;
                    case 7:                         // BTFSS
                        if (rd_f(pcore,f)&(1<<b))
                            {pcore->ir=0; dis=1;}   // Discard prefetched
                        break;
                    default:
                        switch(ireg>>9) {
                        case 0x1f:                  // ADDLW
                            i=j=pcore->w; i+=k;
                            pcore->w=i&0xff;
                            if (i&0x100) pcore->status|=CF;
                            else pcore->status&=~CF;
                            if (!(i&0xff)) pcore->status|=ZF;
                            else pcore->status&=~ZF;
                            j&=0xf; j+=k&0xf;
                            if (j&0x10) pcore->status|=DCF;
                            else pcore->status&=~DCF;
                            break;
                        case 0x1e:                  // SUBLW
                            j=k; i=(pcore->w);
                            n=j-i; n&=0xff;
                            pcore->w=n;
                            if (j>=i) pcore->status|=CF;
                            else pcore->status&=~CF;
                            if (!n) pcore->status|=ZF;
                            else pcore->status&=~ZF;
                            j&=0xf; i&=0xf;
                            if (j>=i) pcore->status|=DCF;
                            else pcore->status&=~DCF;
                            break;
                        default:
                            switch(ireg>>8){        // 6 bit OPcodes
                            case 0x7:               // ADDWF
                                i=pcore->w; j=rd_f(pcore,f);
                                n=i+j;
                                if (!d) pcore->w=n&0xff;
                                else wr_f(pcore,f,n&0xff);
                                if (n&0x100) pcore->status|=CF;
                                else pcore->status&=~CF;
                                if (!(n&0xff)) pcore->status|=ZF;
                                else pcore->status&=~ZF;
                                i&=0xf; i+=j&0xf;
                                if (i&0x10) pcore->status|=DCF;
                                else pcore->status&=~DCF;
                                break;
                            case 0x5:               // ANDWF
                                tmp=pcore->w; tmp&=rd_f(pcore,f);
                                if (!d) pcore->w=tmp;
                                else wr_f(pcore,f,tmp);
                                if (!tmp) pcore->status|=ZF;
                                else pcore->status&=~ZF;
                                break;
                            case 0x9:               // COMF
                                tmp=~rd_f(pcore,f);
                                if (!d) pcore->w=tmp;
                                else wr_f(pcore,f,tmp);
                                if (!tmp) pcore->status|=ZF;
                                else pcore->status&=~ZF;
                                break;
                            case 0x3:               // DECF
                                tmp=rd_f(pcore,f)-1;
                                if (!d) pcore->w=tmp;
                                else wr_f(pcore,f,tmp);
                                if (!tmp) pcore->status|=ZF;
                                else pcore->status&=~ZF;
                                break;
                            case 0xb:               // DECFSZ
                                tmp=rd_f(pcore,f)-1;
                                if (!d) pcore->w=tmp;
                                else wr_f(pcore,f,tmp);
                                if (!tmp) {pcore->ir=0; dis=1;} // Discard prefetched
                                break;
                            case 0xa:               // INCF
                                tmp=rd_f(pcore,f)+1;
                                if (!d) pcore->w=tmp;
                                else wr_f(pcore,f,tmp);
                                if (!tmp) pcore->status|=ZF;
                                else pcore->status&=~ZF;
                                break;
                            case 0xf:               // INCFSZ
                                tmp=rd_f(pcore,f)+1;
                                if (!d) pcore->w=tmp;
                                else wr_f(pcore,f,tmp);
                                if (!tmp) {pcore->ir=0; dis=1;} // Discard prefetched
                                break;
                            case 0x4:               // IORWF
                                tmp=pcore->w; tmp|=rd_f(pcore,f);
                                if (!d) pcore->w=tmp;
                                else wr_f(pcore,f,tmp);
                                if (!tmp) pcore->status|=ZF;
                                else pcore->status&=~ZF;
                                break;
                            case 0x8:               // MOVF
                                tmp=rd_f(pcore,f);
                                if (!d) pcore->w=tmp;
                                else wr_f(pcore,f,tmp);
                                if (!tmp) pcore->status|=ZF;
                                else pcore->status&=~ZF;
                                break;
                            case 0xd:               // RLF
                                i=rd_f(pcore,f); i<<=1;
                                if (pcore->status&CF) i++;
                                if (i&0x100) pcore->status|=CF;
                                else pcore->status&=~CF;
                                if (!d) pcore->w=i&0xff;
                                else wr_f(pcore,f,i&0xff);
                                break;
                            case 0xc:               // RRF
                                i=rd_f(pcore,f);
                                if (pcore->status&CF) i|=0x100;
                                if (i&0x1) pcore->status|=CF;
                                else pcore->status&=~CF;
                                i>>=1;
                                if (!d) pcore->w=i;
                                else wr_f(pcore,f,i);
                                break;
                            case 0x2:               // SUBWF
                                i=(pcore->w);
                                j=rd_f(pcore,f);
                                n=j-i; n&=0xff;
                                if (!d) pcore->w=n;
                                else wr_f(pcore,f,n&0xff);
                                if (j>=i) pcore->status|=CF;
                                else pcore->status&=~CF;
                                if (!n) pcore->status|=ZF;
                                else pcore->status&=~ZF;
                                j&=0xf; i&=0xf;
                                if (j>=i) pcore->status|=DCF;
                                else pcore->status&=~DCF;
                                break;
                            case 0xe:               // SWAPF
                                i=tmp=rd_f(pcore,f);
                                tmp<<=4; i>>=4; tmp|=i;
                                if (!d) pcore->w=tmp;
                                else wr_f(pcore,f,tmp);
                                break;
                            case 0x6:               // XORWF
                                tmp=pcore->w; tmp^=rd_f(pcore,f);
                                if (!d) pcore->w=tmp;
                                else wr_f(pcore,f,tmp);
                                if (!tmp) pcore->status|=ZF;
                                else pcore->status&=~ZF;
                                break;
                            case 0x39:              // ANDLW
                                pcore->w &= k;
                                if (!pcore->w) pcore->status|=ZF;
                                else pcore->status&=~ZF;
                                break;
                            case 0x38:              // IORLW
                                pcore->w |= k;
                                if (!pcore->w) pcore->status|=ZF;
                                else pcore->status&=~ZF;
                                break;
                            case 0x3a:              // XORLW
                                pcore->w ^= k;
                                if (!pcore->w) pcore->status|=ZF;
                                else pcore->status&=~ZF;
                                break;
                            default:                // 7 bit OPcodes
                                switch(ireg>>7) {
                                case 0x3:           // CLRF
                                    wr_f(pcore,f,0);
                                    pcore->status|=ZF;
                                    break;
                                case 0x2:           // CLRW
                                    pcore->w=0;
                                    pcore->status|=ZF;
                                    break;
                                case 0x1:           // MOVWF
                                    wr_f(pcore,f,pcore->w);
                                    break;
                                default:            // 14 bit OPcodes
                                    switch(ireg) {
                                    case 0x63:      // SLEEP
                                        pcore->sleep=1;
                                        pcore->status&=~NPDF;
                                        pcore->wdt=0;
                                        if(pcore->option&0x80)
                                                pcore->prescaler=0;
                                        break;
                                    case 0x64:      // CLRWDT
                                        pcore->wdt=0;
                                        pcore->status|=NTOF|NPDF;
                                        if(pcore->option&0x80)
                                                pcore->prescaler=0;
                                        break;
                                    case 0x9:       // RETFIE
                                        pcore->intcon|=0x80; //enable GIE
                                    case 0x8:       // RETURN
                                        dis=1;
                                        pcore->ir=0;  // Discard prefetched
                                        // old stack level
                                        pcore->sp=(pcore->sp-1)&7;
                                        pcwr=1;
                                        break;
                                    case 0x65: // TRIS PORTA
                                        wr_f(pcore,0x85,pcore->w);
                                        break;
                                    case 0x66: // TRIS PORTB
                                        wr_f(pcore,0x86,pcore->w);
                                        break;
                                    case 0x62: // OPTION
                                        wr_f(pcore,0x81,pcore->w);
                                        break;
                                    }
                                // All the remaining OPcodes are executed
                                // As NOP
                                }
                            }
                        }
                    }
                }

                // Other Hardware Checks:
chkhw:          // Increment profiling counter. This is not a real counter
                // but our simulated time variable.
                pcore->Clocks++;

                // Update F84 counters: Prescaler & TMR0
                b=rd_f(pcore,5)&0x10;   // check for PA4 edge
                if ((pcore->option&0x10)) b^=0x10;
                if (((~opa)&0x10) & b) k=1; else k=0;
                opa=b;
                if ((!(pcore->option&0x20) && !pcore->sleep) ||
                    ((pcore->option&0x20) && k))  { // TS0=0: main clock
                    if (!(pcore->option&0x8)) {     // PSA: prescaled
                        pcore->prescaler++;
                        if (pcore->sync)pcore->sync--;
                        else {
                          if ((~pcore->prescaler)&opr&(1<<(pcore->option&7)))
                                pcore-> tmr0++;
                          opr=pcore->prescaler;
                        }
                    } else {
                        if (pcore->sync)pcore->sync--;
                        else pcore->tmr0++;
                    }
                    //Update timer overflow flag
                    if ((~pcore->tmr0)&otmr0&0x80) pcore->intcon|=4;
                    otmr0=pcore->tmr0;
                }
                // Eeprom write timming
                if (eetim) {
                    if (!(--eetim)) {
                        pcore->eeprom[eeadr]=eedata;
                        pcore->eecon1&=~2; pcore->eecon1|=0x10;
                    }
                }
                // Check for interrupts
                k=0;
                if ((pcore->intcon&0x24)==0x24) k=1; // Timer int
                b=rd_f(pcore,6);
                if ((b^opb)&0xf0) pcore->intcon|=1;  // Portb change
                if ((pcore->intcon&0x9)==0x9) k=1;
                if (pcore->option&0x40) b^=1;
                if ((b&1) && !(opb&1)) pcore->intcon|=2;  // RB0 int
                if ((pcore->intcon&0x12)==0x12) k=1;
                opb=b;
                // Eeprom write interrupt
                if ((pcore->intcon&0x40) && (pcore->eecon1&0x10)) k=1;
                // If an interrupt takes place the F84 awakes:
                if (k) pcore->sleep=0;
                // Check the Global Interrupt Enable bit
                if ((!dis) && (pcore->intcon & 0x80) && k) {
                    trace_interrupt(pcore);
                    // Interrupt Routine Call:
                    pcore->intcon&=0x7f;            // Disable GIE
                    pcore->sp=(pcore->sp+1)&7;      // new stack level
                    pcore->pc[pcore->sp]=4;         // Goto 4
                    pcore->ir=0;                    // Discard prefetched
                    pcwr=1;
                }

                // Watchdog timer: 18 ms nominal delay
                if (pcore->config&4) {
                    pcore->wdt++;
                    if (pcore->wdt>=18*(pcore->freq/1000)){
                        // WatchDog time expired: Reset CPU or increment
                        // Prescaler depending on OPTION settings.
                        pcore->wdt=0;
                        if (pcore->option&8) {
                            pcore->prescaler++;
                            // Reset on prescaler overflow:
                            if (pcore->prescaler&(1<<(pcore->option&7))) {
                                pcore->status&=~NTOF;
                                f84_reset(pcore,WDTR);
                                pcwr=1;
                            }
                        } else {
                            pcore->status&=~NTOF;
                            f84_reset(pcore,WDTR);
                            pcwr=1;
                        }
                    }
                }

                // Increment PC if no PC write/change is taking place
                if (!pcwr && !pcore->sleep) pcore->pc[pcore->sp]++;

                // Sound emulation: see "sound.c" for details
#ifdef SOUND
                sound_tic(pcore);
#endif
        }
}

